@@ -1,12 +1,9 @@
 package Plack::Middleware::Deflater;
 use strict;
 use 5.008001;
-our $VERSION = '0.09';
+our $VERSION = '0.10';
 use parent qw(Plack::Middleware);
 use Plack::Util::Accessor qw( content_type vary_user_agent);
-
-use IO::Compress::Deflate;
-use IO::Compress::Gzip;
 use Plack::Util;
 
 sub prepare_app {
@@ -73,10 +70,8 @@ sub call {
         }
 
         my $encoder;
-        if ($encoding eq 'gzip') {
-            $encoder = sub { IO::Compress::Gzip->new($_[0]) };
-        } elsif ($encoding eq 'deflate') {
-            $encoder = sub { IO::Compress::Deflate->new($_[0]) };
+        if ($encoding eq 'gzip' || $encoding eq 'deflate') {
+            $encoder = Plack::Middleware::Deflater::Encoder->new($encoding);
         } elsif ($encoding ne 'identity') {
             my $msg = "An acceptable encoding for the requested resource is not found.";
             @$res = (406, ['Content-Type' => 'text/plain'], [ $msg ]);
@@ -86,31 +81,21 @@ sub call {
         if ($encoder) {
             $h->set('Content-Encoding' => $encoding);
             $h->remove('Content-Length');
-            my $buf;
-            my $state = 0; # 0: start 1: readed first compressed chunk 2: done all
-            my $compress = $encoder->(\$buf);
-            my $bodybuf = '';
+
+            # normal response
+            if ( $res->[2] && ref($res->[2]) && ref($res->[2]) eq 'ARRAY' ) {
+                my $buf = '';
+                foreach (@{$res->[2]}) {
+                    $buf .= $encoder->print($_) if defined $_;
+                }
+                $buf .= $encoder->close();
+                $res->[2] = [$buf];
+                return;
+            }
+
+            # delayed or stream
             return sub {
-                my $chunk = shift;
-                return if $state == 2;
-                unless (defined $chunk) {
-                    $state = 2;
-                    $compress->close;
-                    $bodybuf .= $buf if defined $buf;
-                    return $bodybuf;
-                }
-                $compress->print($chunk);
-                if (defined $buf) {
-                    $bodybuf .= $buf;
-                    undef $buf;
-                    if ( $state == 0 ) { 
-                        #buffering a first chunk. It contains only the gzip header.
-                        $state = 1;
-                        return '';
-                    }
-                    return substr($bodybuf, 0, length($bodybuf), '');
-                }
-                return '';
+                $encoder->print(shift);
             };
         }
     });
@@ -118,6 +103,72 @@ sub call {
 
 1;
 
+package Plack::Middleware::Deflater::Encoder;
+
+use strict;
+use warnings;
+use Compress::Zlib;
+
+use constant GZIP_MAGIC => 0x1f8b;
+
+sub new {
+    my $class = shift;
+    my $encoding = shift;
+    my ($encoder,$status) = $encoding eq 'gzip'
+        ? deflateInit(-WindowBits => -MAX_WBITS())
+        : deflateInit(-WindowBits => MAX_WBITS());
+    die 'Cannot create a deflation stream' if $status != Z_OK;
+    
+    bless {
+        header => 0,
+        closed => 0,
+        encoding => $encoding,
+        encoder => $encoder,
+        crc => crc32(undef),
+        length => 0,
+    }, $class;
+}
+
+sub print : method {
+    my $self = shift;
+    return if $self->{closed};
+    my $chunk = shift;
+    if ( ! defined $chunk ) {
+        my ($buf,$status) = $self->{encoder}->flush();
+        die "deflate failed: $status" if ( $status != Z_OK );
+        if ( !$self->{header} && $self->{encoding} eq 'gzip' ) {
+            $buf = pack("nccVcc",GZIP_MAGIC,Z_DEFLATED,0,time(),0,$Compress::Raw::Zlib::gzip_os_code) . $buf
+        }
+        $buf .= pack("LL",$self->{length},$self->{crc}) if $self->{encoding} eq 'gzip';
+        $self->{closed} = 1;
+        return $buf;
+    }
+
+    my ($buf,$status) = $self->{encoder}->deflate($chunk);
+    die "deflate failed: $status" if ( $status != Z_OK );
+    $self->{length} += length $chunk;
+    $self->{crc} = crc32($_[0],$self->{crc});
+    if ( length $buf ) {
+        if ( !$self->{header} && $self->{encoding} eq 'gzip' ) {
+            $buf = pack("nccVcc",GZIP_MAGIC,Z_DEFLATED,0,time(),0,$Compress::Raw::Zlib::gzip_os_code) . $buf
+        }
+        $self->{header} = 1;
+        return $buf;
+    }
+    return '';
+}
+
+sub close : method {
+    $_[0]->print(undef);
+}
+
+sub closed {
+    $_[0]->{closed};
+}
+
+1;
+
+
 __END__
 
 =head1 NAME
